
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; MODULE      : menu.scm
;; DESCRIPTION : routines for generating menus
;; COPYRIGHT   : (C) 2002  Joris van der Hoeven, David Allouche
;;
;; This software falls under the GNU general public license and comes WITHOUT
;; ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for details.
;; If you don't have this file, write to the Free Software Foundation, Inc.,
;; 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Menu utilities
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (make-menu-empty) (widget-harray () -1))
(define (make-menu-error . args)
  (apply tm-display-error args)
  (widget-text "None" #f ""))

(define (pixel) 256) ; TODO: length arithmetic

(define (make-menu-button but lab)
  (if (tuple? lab 'balloon 2)
      (widget-balloon but
		      (widget-text (caddr lab) #t (get-input-language)))
      but))

(define (make-menu-make-widget p active?)
  (widget-promise (list 'make-menu-widget (list 'quote p) active?)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Automatic check marks
;; Should be moved to corresponding routines later
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (set-check-mark! proc mark pred)
  (set-procedure-property! proc 'check-mark (list mark pred)))
(define (check-mark proc)
  (procedure-property proc 'check-mark))

(set-check-mark! make-line-with "o" test-env?)
(set-check-mark! make-with "o" test-env?)
(set-check-mark! init-env "*" test-init?)
(set-check-mark! init-default "*" test-default?)

(define (test-shrinking-factor? n) (= (get-shrinking-factor) n))
(set-check-mark! toggle-full-screen-mode "v" full-screen?)
(set-check-mark! toggle-visible-header "v" visible-header?)
(set-check-mark! toggle-visible-icon-bar "v" visible-icon-bar?)
(set-check-mark! toggle-visible-footer "v" visible-footer?)
(set-check-mark! set-shrinking-factor "*" test-shrinking-factor?)

(define (test-preamble?) (string=? (get-env "preamble") "true"))
(define (test-page-medium? s) (string=? (get-env "page medium") s))
(define (test-page-type? s) (string=? (get-env "page type") s))
(define (test-page-orientation? s) (string=? (get-env "page orientation") s))
(define (visible-header-and-footer?)
  (equal? "true" (get-env "show header and footer")))
(set-check-mark! toggle-preamble "v" test-preamble?)
(set-check-mark! set-page-medium "*" test-page-medium?)
(set-check-mark! set-page-type "*" test-page-type?)
(set-check-mark! set-page-orientation "*" test-page-orientation?)
(set-check-mark! toggle-visible-header-and-footer
		 "v" visible-header-and-footer?)
; FIXME: check init-as-on-paper; should become a persistant property

(define (selection-test-import? s) (string=? s (selection-get-import)))
(define (selection-test-export? s) (string=? s (selection-get-export)))
(define (session-test-math-input? b?) (if (session-math-input?) b? (not b?)))
(set-check-mark! selection-set-import "*" selection-test-import?)
(set-check-mark! selection-set-export "*" selection-test-export?)
(set-check-mark! session-use-math-input "*" session-test-math-input?)
(set-check-mark! toggle-insertion-position "v" test-insertion-position?)

(define (test-cell-mode? s) (string=? (get-cell-mode) s))
(define (table-test-halign? s) (string=? (table-get-halign) s))
(define (table-test-valign? s) (string=? (table-get-valign) s))
(define (cell-test-halign? s) (string=? (cell-get-halign) s))
(define (cell-test-valign? s) (string=? (cell-get-valign) s))
(define (cell-test-hmode? s) (string=? (cell-get-hmode) s))
(define (cell-test-vmode? s) (string=? (cell-get-vmode) s))
(define (cell-test-background? s) (string=? (cell-get-background) s))
(define (cell-test-vcorrect? s) (string=? (cell-get-vcorrect) s))
(define (cell-test-hyphen? s) (string=? (cell-get-hyphen) s))
(set-check-mark! set-cell-mode "*" test-cell-mode?)
(set-check-mark! table-set-halign "*" table-test-halign?)
(set-check-mark! table-set-valign "*" table-test-valign?)
(set-check-mark! toggle-table-hyphen "v" table-hyphen?)
(set-check-mark! cell-set-halign "o" cell-test-halign?)
(set-check-mark! cell-set-valign "o" cell-test-valign?)
(set-check-mark! cell-set-hmode "o" cell-test-hmode?)
(set-check-mark! cell-set-vmode "o" cell-test-vmode?)
(set-check-mark! cell-set-background "o" cell-test-background?)
(set-check-mark! cell-set-vcorrect "o" cell-test-vcorrect?)
(set-check-mark! cell-set-hyphen "o" cell-test-hyphen?)
(set-check-mark! cell-toggle-multi-paragraph "v" cell-multi-paragraph?)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Menu items
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (menu-label? p)
"p --> <boolean>
Does p match the syntactic structure of a menu label definition?

Label with a popup balloon. The <string> is the balloon text.
    <label> :: (balloon <label> <string>)

Label <string> drawn in black text of an arbitrary font.
    <label> :: (text <font desc> <string>)

Example default values are: family=\"roman\", class=\"mr\", series=\"medium\",
shape=\"normal\", size=10, dpi=600.
    <font desc> :: ([family [class [series [shape [size [dpi]]]]]])

Simple menu label, its display style is controlled by tt? and active? flags.
    <label> :: <string>

Pixmap menu label, the <string> is the name of the pixmap.
    <label> :: (icon <string>)"

  (or (and (tuple? p 'balloon 2)	; (balloon <label> "balloon text")
	   (menu-label? (cadr p))
	   (string? (caddr p)))
      (and (tuple? p 'text 2)		; (text <font desc> "text")
	   (tuple? (cadr p))
	   (string? (caddr p)))
      (string? p)			; "text"
      (and (tuple? p 'icon 1)		; (icon "name.xpm")
	   (string? (cadr p)))))


(define (make-menu-label p active? . opt)
  "p, active?, tt? = #f --> <widget>
Assuming that (menu-label? p) is true, evaluate p to a widget.
See the menu-label? docstring for the syntax of p"

  (let ((tt? (and (not (null? opt)) (car opt))))
    (cond ((tuple? p 'balloon 2)        ; (balloon <label> "balloon text")
	   (make-menu-label (cadr p) active? tt?))
	  ((tuple? p 'text 2)		; (text <font desc> "text")
	   (widget-box (cadr p) (caddr p) #t #t))
	  ((string? p)			; "text"
	   (widget-menu-text p
			     (color (if active? "black" "dark grey"))
			     (get-input-language)
			     tt?))
	  (else				; (icon "name.xpm")
	   (widget-xpm (cadr p) #t)))))


(define (make-menu-item p active?)
  "<menu-item> :: (<label> [...] [<shortcut>] [<check>] <action>)"

  (define (make-button active? label short command)
    (if (string=? short "")
	(widget-command-button-1 (make-menu-label label active?)
				 command #f)
	(widget-command-button-2 (make-menu-label label active?)
				 (make-menu-label short active? #t)
				 command)))

  (define (make-shortcut label action opt-key)
    (cond (opt-key (cdr opt-key))
	  ((pair? label) "")
	  (else (get-inverse-binding action))))

  (define (make-check opt-check action)
    (or opt-check
	(and (pair? action)
	     (let ((proc (eval (car action))))
	       (and (procedure? (eval (car action)))
		    (let ((prop (procedure-property proc 'check-mark)))
		      (and prop
			   `(check ,(car prop)
				   (,(cadr prop) . ,(cdr action))))))))))

  (define (make-label label opt-dots opt-check)
    (if (pair? label) label
	(string-append
	 (if (and opt-check
		  (eval-bool-or-false (caddr opt-check)))
	     (string-append (cadr opt-check) "#")
	     "")
	 label (if opt-dots "..." ""))))
  
  (let ((action (list-ref p (1- (length p))))
	(label  (car p)))
    (apply-with
     (list->alist `((key   ,menu-label?)
		    (dots  ,(lambda (x) (equal? '... x)))
		    (check ,(lambda (x) (equal? 'check (car x))) ,cdr))
		  (list-head (cdr p) (- (length p) 2)))
     (lambda (error? opts)
       (if error? (make-menu-error
		   "invalid menu item attribute at position " (1+ opts)
		   " in " (object->string p))
	   (let ((opt-key (assq 'key opts))
		 (opt-dots (assq 'dots opts))
		 (opt-check (make-check (assq 'check opts) action)))
	     (make-button
	      active?
	      (make-label label opt-dots opt-check)
	      (make-shortcut label action opt-key)
	      (server-command (if active? action 'noop)))))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Menu bodies
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (make-menu-bad-format p enbl?)
  (make-menu-error "menu has bad format in " (object->string p)))

;; ENTRY POINT
(define (make-menu-widget p enbl?)
  "Generates menubar and menus."

  (define (make-vmenu p enbl?) (widget-vmenu (make-menu-items (cdr p) enbl?)))

  (define (make-linked-menu p enbl?)
    (apply-with (safe-eval-list (cadr p))
		(lambda (error? lst)
		  (if (not error?)
		      (make-menu-widget lst enbl?)
		      (make-menu-error
		       "error evaluating link: " (object->string (cadr p))
		       ", result: " (object->string lst))))))

  (define (make-hmenu p enbl?) (widget-harray (make-menu-items p enbl?) -1))
    
;   (set-trace-point! make-vmenu      "case: vmenu")
;   (set-trace-point! make-linked-menu  "case: link")
;   (set-trace-point! make-menu-items "case: hmenu")
  
  (cond ((not (tuple? p)) (make-menu-bad-format p enbl?))
	((tuple? p 'ver-menu) (make-vmenu p enbl?))
	((tuple? p 'link 1)   (make-linked-menu p enbl?))
	(else (make-hmenu p enbl?))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Menu items
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (make-menu-items p enbl?)
  (filter (flatten (make-menu-items* p enbl?))
	  (lambda (x) (not (null? x)))))

(define (make-menu-items* p enbl?)
  "Generate menu items.
  <menu-entry> :: <item-list> | <if-list> | <when-list> | <item-link>"

  (define (make-hsep)
    (widget-separator (* 2 (pixel)) (* 2 (pixel)) #f))
  (define (make-vsep)
    (widget-separator (* 2 (pixel)) (* 2 (pixel)) #t))

  (define (make-item-list enbl? p)
    "<item-list> :: ([<menu-entry> ...])"
    (map (lambda (pp) (make-menu-items* pp enbl?)) p))

  (define (make-items-if enbl? tag pred? . items)
    "<if-list> :: (if <predicate> . <item-list>)"
    (if (eval-bool-or-false pred?) (make-item-list enbl? items) ()))

  (define (make-items-when enbl? tag pred? . items)
    "<when-list> :: (when <predicate> . <item-list>)"
    (let ((enbl? (and enbl? (eval-bool-or-false pred?))))
      (make-item-list enbl? items)))

  (define (make-submenu enbl? tag label . items)
    "<submenu> :: (<gravity> <menu-label> [<menu-entry> ...])
<gravity> :: => | ->"
    (make-menu-button
     ((cond ((equal? tag '=>) widget-pulldown-button-lazy)
	    ((equal? tag '->) widget-pullright-button-lazy))
      (make-menu-label label enbl?)
      (make-menu-make-widget (append! (list 'ver-menu) items)
			     enbl?))
     label))

  (define (make-tile-menu enbl? tag width . items)
    "<tile-menu> :: (tile <integer> [<menu-entry> ...])"
    (widget-tile (make-menu-items items enbl?) width))

  (define (make-linked-menu-items enbl? tag expr)
    "<item-link> :: (link <expr yielding <menu-entry>>)"
    (apply-with (safe-eval-list expr)
		(lambda (error? r)
		  (if (not error?)
		      (make-menu-items* r enbl?)
		      (make-menu-error
		       "error evaluating link: " (object->string expr))))))

  (define (make-menu-entry enbl? p)
    "<menu-entry> :: <menu-item>"
    (make-menu-button (make-menu-item p enbl?)
		      (car p)))

  (define (make-symbol-button enbl? symstring opt-alist)
    (widget-command-button-1
     (widget-box () symstring #t #f)
     (server-command (if (assq 'symobj opt-alist)
			 (list 'insert-object
			       (list 'quote (cdr (assq 'symobj opt-alist))))
			 (list 'insert-string symstring)))
     #f))

  (define (make-menu-symbol enbl? tag symstring . opt)
    "<menu-symbol> :: (symbol <symbol-string> [<symbol-object>] [<shorcut>])
<symbol-object> :: (<symbol-type> <symbol-name>)
<symbol-type> :: left | right | mid | big
<symbol-name> :: <string>"
    (apply-with
     (list->alist (list (list 'symobj pair?) (list 'short string?))
		  opt)
     (lambda (error? opt-alist)
       (if error? (make-menu-error
		   "invalid symbol attribute at position " (+ 2 opt-alist)
		   " in " (list tag symstring . opt))
	   (let ((sh (if (assq 'short opt-alist)
			 (cdr (assq 'short opt-alist))
			 (get-inverse-binding (string->symbol symstring)))))
	     (if (string=? sh "")
		 (make-symbol-button enbl? symstring opt-alist)
		 (widget-balloon
		  (make-symbol-button enbl? symstring opt-alist)
		  (make-menu-label (string-append "Keyboard equivalent: " sh)
				   enbl?))))))))

;   (set-trace-point! make-hsep          "case: ---")
;   (set-trace-point! make-vsep          "case: |")
;   (set-trace-point! make-item-list     "case: list")
;   (set-trace-point! make-items-if      "case: if")
;   (set-trace-point! make-items-when    "case: when")
;   (set-trace-point! make-submenu       "case: =>/->")
;   (set-trace-point! make-tile-menu     "case: tile")
;   (set-trace-point! make-linked-menu-items "case: link")
;   (set-trace-point! make-menu-entry    "case: item")
;   (set-trace-point! make-menu-symbol   "case: symbol")

  (let ((n (if (tuple? p) (length p))))
    (cond ((equal? p '---) (make-hsep))
	  ((equal? p '|) (make-vsep))
	  ((not (tuple? p)) (make-menu-bad-format p enbl?))
	  ((and (>= n 2) (menu-label? (car p)))
	   (make-menu-entry enbl? p))
	  ((or (= n 0) (tuple? (car p)))
	   (make-item-list enbl? p))
	  (else
	   (let ((handler
		  (assq (car p)
			`((if     ,make-items-if 2 *)
			  (when   ,make-items-when 2 *)
			  (->     ,make-submenu 2 * ,menu-label?)
			  (=>     ,make-submenu 2 * ,menu-label?)
			  (tile   ,make-tile-menu 2 * ,integer?)
			  (link   ,make-linked-menu-items 2 2)
			  (symbol ,make-menu-symbol 2 4 ,string?)))))
	     (cond ((not handler)
		    (make-item-list enbl? p))
		   ((and (>=  n (caddr handler))
			 (or (equal? '* (cadddr handler))
			     (<= n (cadddr handler)))
			 (or (null? (cddddr handler))
			     (apply (car (cddddr handler))
				    (list (cadr p)))))
		    (apply (cadr handler) (cons enbl? p)))
		   (else (make-menu-bad-format p enbl?))))))))

; (set-trace-level! make-menu-widget)
; (set-trace-level! make-menu-item)
; (set-trace-level! make-menu-items*)
; (set-trace-level! make-menu-items)
			
(set! make-menu-widget (wrap-catch make-menu-widget))
