
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; MODULE      : tmhtml.scm
;; DESCRIPTION : conversion of TeXmacs trees into Html trees
;; COPYRIGHT   : (C) 2002  Joris van der Hoeven
;;
;; This software falls under the GNU general public license and comes WITHOUT
;; ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for details.
;; If you don't have this file, write to the Free Software Foundation, Inc.,
;; 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Data
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define tmhtml-initialized #f)
(define tmhtml-methods (make-hash-table 500))
(define tmhtml-stdmarkup (make-hash-table 500))
(define tmhtml-with-cmd (make-hash-table 100))
(define tmhtml-env (make-hash-table 1000))

(define tmhtml-methods-def
  '((document tmhtml-document)
    (paragraph tmhtml-paragraph)
    (surround tmhtml-noop)
    (concat tmhtml-concat)
    (format tmhtml-format)
    (hspace vspace* vspace space htab split move resize float
     repeat datoms dlines dpages dbox tmhtml-noop)
    (group left mid right big lprime rprime below above
     lsub lsup rsub rsup frac sqrt wide neg
     tree old_matrix old_table old_mosaic old_mosaic_item
     tformat twith cwith tmarker
     table row cell sub_table tmhtml-noop)

    (assign tmhtml-noop)
    (with tmhtml-with)
    (set reset tmhtml-noop)
    (var_expand expand apply tmhtml-expand)
    (begin end include macro func env eval tmhtml-noop)
    (value tmhtml-expand)
    (arg tmhtml-noop)
    (backup quote delay hold release tmhtml-noop)

    (or xor and not plus minus times over div mod merge length range
     number date translate is_tuple look_up equal unequal less lesseq
     greater greatereq if case while extern authorize tmhtml-noop)
    (inactive symbol latex hybrid tuple collection associate tmhtml-noop)
    (label tmhtml-label)
    (reference tmhtml-noop)
    (pageref tmhtml-noop)
    (write tmhtml-noop)
    (specific tmhtml-specific)
    (hlink tmhtml-hyperlink)
    (action tag meaning tmhtml-noop)
    (switch fold exclusive progressive superposed tmhtml-noop)
    (graphics point line arc bezier tmhtml-noop)
    (postscript tmhtml-postscript)

    (!file tmhtml-file)))

(define tmhtml-stdmarkup-def
  '((chapter (h1))
    (section (h2))
    (subsection (h3))
    (subsubsection (h4))
    (paragraph (h5))
    (subparagraph (h6))
    (itemize (ul))
    (enumerate (ol))
    (description (dl))
    (item (li))
    ;(item* (dt))
    (strong (strong))
    (em (em))
    (tt (tt))
    (name (name))
    (samp (samp))
    (abbr (abbr))
    (dfn (dfn))
    (kbd (kbd))
    (var (var))
    (acronym (acronym))
    (person (person))))

(define tmhtml-with-cmd-def
  '((("font family" "tt") (tt))
    (("font series" "bold") (b))
    (("font shape" "italic") (i))
    (("color" "black") (font (@ ("color" "black"))))
    (("color" "grey") (font (@ ("color" "grey"))))
    (("color" "white") (font (@ ("color" "white"))))
    (("color" "red") (font (@ ("color" "red"))))
    (("color" "blue") (font (@ ("color" "blue"))))
    (("color" "yellow") (font (@ ("color" "black"))))
    (("color" "magenta") (font (@ ("color" "magenta"))))
    (("color" "orange") (font (@ ("color" "orange"))))
    (("color" "green") (font (@ ("color" "green"))))
    (("color" "brown") (font (@ ("color" "brown"))))
    (("paragraph mode" "left") (div (@ ("align" "left"))))
    (("paragraph mode" "justify") (div (@ ("align" "justify"))))
    (("paragraph mode" "center") (center))
    (("paragraph mode" "right") (div (@ ("align" "right"))))
    (("nr columns" "1") (multicol (@ ("cols" "1"))))
    (("nr columns" "2") (multicol (@ ("cols" "2"))))
    (("nr columns" "3") (multicol (@ ("cols" "3"))))
    (("nr columns" "4") (multicol (@ ("cols" "4"))))
    (("nr columns" "5") (multicol (@ ("cols" "5"))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Initialization
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (tmhtml-initialize)
  (if (not tmhtml-initialized)
      (begin
	(set! tmhtml-initialized #t)
	(fill-dictionary tmhtml-methods tmhtml-methods-def)
	(fill-dictionary tmhtml-stdmarkup tmhtml-stdmarkup-def)
	(fill-dictionary tmhtml-with-cmd tmhtml-with-cmd-def)))
  (set! tmhtml-env (make-hash-table 1000)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Manipulation of the environment
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (tmhtml-env-list var)
  (let ((r (hash-ref tmhtml-env var)))
    (if r r '())))

(define (tmhtml-env-get var)
  (let ((val (tmhtml-env-list var)))
    (if (null? val) #f
	(car val))))

(define (tmhtml-math-mode?)
  (equal? (tmhtml-env-get "mode") "math"))

(define (tmhtml-env-set var val)
  (hash-set! tmhtml-env var (cons val (tmhtml-env-list var))))

(define (tmhtml-env-reset var)
  (let ((val (tmhtml-env-list var)))
    (if (not (null? val))
	(hash-set! tmhtml-env var (cdr val)))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Strings
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (tmhtml-token-sub s)
  (cond ((equal? s "less") #\<)
	((equal? s "gtr") #\>)
	((equal? s "box") (list 'Box))
	((equal? s "||") (list '|))
	(else (list (string->symbol s)))))

(define (tmhtml-token l routine)
  (let* ((p (split (cdr l) (lambda (x) (equal? x #\>))))
	 (s (list->string (car p)))
	 (q (if (null? (cadr p)) '() (cdadr p)))
	 (r (routine q)))
    (cons (tmhtml-token-sub s) r)))

(define (tmhtml-text-sub head l)
  (append (string->list head) (tmhtml-text-list (cdr l))))

(define (tmhtml-text-list l)
  (if (null? l) l
      (let ((c (car l)))
	(cond ((equal? c #\<) (tmhtml-token l tmhtml-text-list))
	      ((equal? c #\21) (tmhtml-text-sub "''" l))
	      ((equal? c #\22) (tmhtml-text-sub ",," l))
	      ((equal? c #\25) (tmhtml-text-sub "--" l))
	      ((equal? c #\26) (tmhtml-text-sub "---" l))
	      ((equal? c #\337) (tmhtml-text-sub "SS" l))
	      ((equal? c #\377) (tmhtml-text-sub "" l))
	      (else (cons c (tmhtml-text-list (cdr l))))))))

(define (tmhtml-verb-list l)
  (if (null? l) l
      (let ((c (car l)))
	(if (equal? c #\<)
	    (let ((r (tmhtml-token l tmhtml-verb-list)))
	      (if (char? (car r)) r (cdr r)))
	    (cons c (tmhtml-verb-list (cdr l)))))))

(define (tmhtml-string-produce l)
  (if (null? l) l
      (if (char? (car l))
	  (let ((p (split l (lambda (x) (not (char? x))))))
	    (cons (list->string (car p))
		  (tmhtml-string-produce (cadr p))))
	  (cons (car l) (tmhtml-string-produce (cdr l))))))

(define (tmhtml-text-string s)
  (let* ((l (string->list s))
	 (t (tmhtml-text-list l))
	 (r (tmhtml-string-produce t)))
    (tex-concat r)))

(define (tmhtml-verb-string s)
  (let* ((l (string->list s))
	 (t (tmhtml-verb-list l))
	 (r (tmhtml-string-produce t)))
    (tex-concat r)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Simple text
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (tmhtml-noop l) "")

(define (tmhtml-file l)
  (let* ((doc (car l))
	 (styles (cdadr l))
	 (lang (caddr l))
	 (tmpath (cadddr l)))
    (list 'html (cons* 'body '(@ ("align" "justify"))
		       (tmhtml-list (list doc))))))

(define (tmhtml-document l)
  (cons '!concat (map-unary
		 (lambda (x) (list 'p x))
		 (tmhtml-list l))))

(define (tmhtml-paragraph-breaks l)
  (if (or (null? l) (null? (cdr l))) l
      (cons* (car l) '(br) (tmhtml-paragraph-breaks (cdr l)))))

(define (tmhtml-paragraph l)
  (cons '!concat (tmhtml-paragraph-breaks (tmhtml-list l))))

(define (tmhtml-concat l)
  (cons '!concat (tmhtml-list l)))

(define (tmhtml-format l)
  (let ((s (car l)))
    (cond ((equal? s "line break") '(br))
	  ((equal? s "new line") '(p))
	  ((equal? s "next line") '(p))
	  (else ""))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Local and global environment changes
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (tmhtml-get-with-cmd var val)
  (if (equal? var "font size")
      (let ((x (* (string->number val) 10)))
	(cond ((< x 1) #f)
	      ((< x 5.5) (font (@ size -4)))
	      ((< x 6.5) (font (@ size -3)))
	      ((< x 7.5) (font (@ size -2)))
	      ((< x 9.5) (font (@ size -1)))
	      ((< x 11.5) (font (@ size +0)))
	      ((< x 13.5) (font (@ size +1)))
	      ((< x 15.5) (font (@ size +2)))
	      ((< x 18.5) (font (@ size +3)))
	      ((< x 22.5) (font (@ size +4)))
	      ((< x 50) (font (@ size +5)))
	      (else #f)))
      (hash-ref tmhtml-with-cmd (list var val))))

(define (tmhtml-with-one var val arg)
  (let ((w (tmhtml-get-with-cmd var val)))
    (cond (w (append w (list arg)))
	  (else arg))))

(define (tmhtml-with l)
  (if (null? l) ""
      (if (null? (cdr l)) (tmhtml (car l))
	  (let ((var (force-string (car l)))
		(val (force-string (cadr l)))
		(next (cddr l)))
	    (tmhtml-env-set var val)
	    (let ((r (tmhtml-with-one var val (tmhtml-with next))))
	      (tmhtml-env-reset var)
	      r)))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Other primitives
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (tmhtml-label l)
  `(a (@ ("name" ,(force-string (car l))))))

;(define (tmhtml-reference l)
;  (list 'ref (force-string (car l))))

;(define (tmhtml-pageref l)
;  (list 'pageref (force-string (car l))))

(define (tmhtml-specific l)
  (if (equal? (car l) "html") (tmhtml (cadr l)) ""))

(define (tmhtml-hyperlink l)
  `(a (@ href ,(force-string (car l)))
      ,(cons 'concat! (tmhtml-list (cdr l)))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; The main conversion routines
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(define (tmhtml-apply key args)
  (let ((n (length args))
	(r (hash-ref tmhtml-methods key))
	(e (hash-ref tmhtml-stdmarkup key)))
    (cond (r ((eval r) args))
	  (e (append e (tmhtml-list args)))
	  (else ""))))

(define (tmhtml-expand l)
  (tmhtml-apply (string->symbol (car l)) (cdr l)))

(define (tmhtml-list-simplify l)
  (if (null? l) l
      (let ((t (tmhtml-list-simplify (cdr l))))
	(cond ((equal? (car l) "") t)
	      ;((func? (car l) '!concat) (append (cdar l) t))
	      (else (cons (car l) t))))))

(define (tmhtml-list l)
  (tmhtml-list-simplify (map-unary tmhtml l)))

(define (tmhtml x)
  (if (string? x) (tmhtml-text-string x)
      (tmhtml-apply (car x) (cdr x))))

(define (tmhtml-produce x)
  (tmhtml-initialize)
  (htmlout-produce (tmhtml (tmpre-produce x))))
